/*
    SR4 Dice is a Dashboard widget for Mac OS X that rolls dice for the
    Shadowrun, Fourth edition roleplaying game by FanPro LLC.

    Copyright (C) 2005  Gurth <gurth@xs4all.nl>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// SR4 Dice CODE VERSION 1.11

function startWidget() {
    /*
      Called by the onload attribute of the widget's BODY tag to do whatever
      initializations need to be done.
     */

    // Stuff that only happens when we're running in Dashboard, not a web
    // browser
    if (window.widget) {
	widget.onshow = showWidget()
    }
}

function showWidget(){
    /*
      Makes sure the widget is properly set-up.
     */

    // Get stored preferences
    var numberOfDice = widget.preferenceForKey("numberOfDice")

    // Set number of dice if no preference exists yet
    if (!numberOfDice) numberOfDice = 4

    // Set visible number of dice to retrieved preference
    visibleDice(numberOfDice)

    // Localize the widget
    languageSetup()
}


function setNumberOfDice(change) {
    /*
      Used by the up and down arrows to increment or decrement the
      numberOfDice text box.
     */

    // Get the current number of dice and change it up or down as needed
    var n = getNumber(document.controls.numberOfDice.value) + change

    // make sure there is at least one die
    if (n < 1) n = 1

    // Set the number of dice displayed in the widget to the new number
    document.controls.numberOfDice.value = n

    // Draw the appropriate number of dice icons
    visibleDice(n)
}

function visibleDice(numberOfDice) {
    /*
      Blanks out unneeded dice icons in the widget display.
     */

    // Check if an argument was passed to the function
    if (numberOfDice)
	// If so, update the text box to show the number of dice we want
	document.controls.numberOfDice.value = numberOfDice
    else
	// If not, display the number of dice the user set in the text box
	numberOfDice = document.controls.numberOfDice.value

    // Make sure we have a legal number
    numberOfDice = getNumber(numberOfDice)

    // Roll all dice to make sure any newly-displayed ones don't all
    // show the same number
    rollDice()

    // If we're running in Dashboard, save current number of dice to preferences
    if (window.widget) widget.setPreferenceForKey(numberOfDice, "numberOfDice")

    // Hide and display the appropriate number of icons
    for (var n = 0; n < document.die.length; n++) {
	if (n < numberOfDice)
	    document.die[n].style.display = "inline" // Display icon
	else
	    document.die[n].style.display = "none" // Hide icon
    }

}

function getNumber(n) {
    /*
      This makes sure the number passed to it is a positive integer. We can't
      be rolling a negative number of dice, can we?
     */

    n = Math.ceil(Math.abs(n))
    return n
}

function rollDice() {
    /*
      Function to do the actual dicerolling and setting of icons.
     */

    // An array to store the rolls. This only needs to be as big as the
    // number of dice to be rolled.
    var rolls = new Array(Math.ceil(document.controls.numberOfDice.value))

    // A counter for the number of hits achieved (a die is a hit if it
    // rolls 5 or 6 -- see SR4 rulebook (sorry, no page ref as I don't
    // actually have a hardcopy yet :)
    var hits = 0

    // And another counter for the number of 1s rolled, in order to check
    // for glitches
    var ones = 0

    // Do the roll as many times as needed
    for (var n = 0; n < rolls.length; n++) {

	// Get a random number from 1-6, see function description below
	rolls[n] = random(1, 6)

	// Check if the latest roll is a hit, and if so, increment hits
	if (rolls[n] >= 5) hits++
	// If the roll is not a hit, check if it's a 1 instead
	else if (rolls[n] == 1) ones++
    }

    // Sort rolls from lowest to highest and then reverse order (so hits are
    // at the beginning instead of at the end -- makes them easier to spot)
    rolls.sort().reverse() 

    // Check if we're rolling more than the number of dice that can be displayed
    // in the widget.
    if (rolls.length > document.die.length)
	var dieLength = document.die.length
    else
	var dieLength = rolls.length

    // This loop goes through the array and sets the dice icons to display
    // the appropriate images. These are named 1.png, 2.png, etc.
    for (n = 0; n < dieLength; n++)
	document.die[n].src = rolls[n].toString() + ".png"

    // call function to display the number of hits rolled
    outputHits(hits, ones, rolls.length)
}

function outputHits(hits, ones, numberOfDice) {
    /*
      This updates the display of the number of hits rolled and inform of any
      glitches that may have occurred.
     */

    // Check if there are any hits at all
    if (!hits)
	// If not, say so
	var h = getLocalizedString("No Hits!")
    else if (hits == 1)
	// One hit
	var h = getLocalizedString("1 Hit!")
    else
	// Every other option is plural
	{
	    // Need to replace the "#" in localized string by actual number
	    // of hits
	    var h = getLocalizedString("# Hits!")
	    var pos = h.indexOf("#")
	    h = h.slice(0, pos) + hits.toString() + h.slice(pos + 1)
	}

    // Now check for a glitch (this is when half or more of the dice are 1s;
    // sorry, again no page ref yet :)
    if (ones >= (numberOfDice / 2)) {
	// If a glitch occurs, check if any hits were rolled
	if (hits) {
	    // If yes, then it's just a normal glitch, so say so
	    h = h + " " + getLocalizedString("Glitch!")
	    }
        else {
	    // With no hits, it's a critical glitch instead
	    h = getLocalizedString("Critical Glitch!")
	    }
    }

    // update the actual display
    document.getElementById("numberOfHits").innerHTML = h

}


//======================================================================

/*
  The following is a random number generator that I took from Erik
  stergaard's web site at http://www.danbbs.dk/~erikoest/rng.htm -- all
  the code in this section came straight from there, so hopefully he
  doesn't object to this :)
  The reason for doing this, instead of using the built-in Math.random()
  Javascript function, is because Dashboard (at least in 10.4.2) does not
  properly seed it -- every time your restart the widget, you get the
  exact same sequence of dice rolls: 1,1,3,5; 1,2,4,5; 3,4,5,6; and so
  on. Thus, this solution.
 */


var genRandNo     // Stores Generated random number.
var uslrn = 0;    // Stores User Selected Lower Range Number.
var usurn = 0;    // Stores User Selected Upper Range Number.

var rand = new RandomNumberGenerator();
genRandNo = random(uslrn, usurn);

// The RandomNumberGenerator (RNG) with the function 
// 'NextRandomNumber()' and the function 
// 'RandomNumberGenerator()' is an implementation of 
// the Park-Miller algorithm. (See 'Random Number Generators: Good 
// Ones Are Hard to Find', by Stephen K. Park and Keith W. Miller, 
// Communications of the ACM, 31(10):1192-1201, 1988.) The JScript 
// version was written by David N. Smith of IBM's T. J. Watson 
// Research Center. Mr. Smith notes that his version has not been 
// subjected to the rigorous testing required of a mission-critical RNG.

// You might have noticed that JScript's Math object includes 
// a built-in random() method. The version presented here should 
// work as well as, if not better than, the built-in implementations, 
// and will work uniformly on all platforms.

function NextRandomNumber()  {
  var hi   = this.seed / this.Q;
  var lo   = this.seed % this.Q;
  var test = this.A * lo - this.R * hi;
  if (test > 0)
    this.seed = test;
  else
    this.seed = test + this.M;
  return (this.seed * this.oneOverM);
}

function RandomNumberGenerator() {
  var d = new Date();
  this.seed = 2345678901 +
    (d.getSeconds() * 0xFFFFFF) +
    (d.getMinutes() * 0xFFFF);
  this.A = 48271;
  this.M = 2147483647;
  this.Q = this.M / this.A;
  this.R = this.M % this.A;
  this.oneOverM = 1.0 / this.M;
  this.next = NextRandomNumber;
  return this;
}

function random(lrn, urn) {
  // Random LowerRange Number (lrn)
  // Random UpperRange Number (urn)
  // return Math.round((urn - lrn + 1) * rand.next() + lrn);
  return Math.floor((urn - lrn + 1) * rand.next() + lrn);
}


//======================================================================
/*
  Stuff required for localization is here. All this is basically just
  Apple-supplied code.
*/

function getLocalizedString (key)
{
    try {
        var ret = localizedStrings[key];
        if (ret === undefined)
            ret = key;
        return ret;
    } catch (ex) {}
    return key;
}

function languageSetup()
{
    document.getElementById("rollButtonText").innerHTML = getLocalizedString("Roll");
    document.getElementById("backBlurb").innerHTML = getLocalizedString("backBlurb");
    document.getElementById("doneButton").innerHTML = getLocalizedString("Done");
}

//======================================================================
/*
  The following is nearly all Apple-supplied code for stuff like the flip
  rollie and showing the widget's back, plus a few other simple odds and ends.
 */


function hideBack() {
	var front = document.getElementById("front");
	var back = document.getElementById("back");

	if (window.widget)
		widget.prepareForTransition("ToFront");

	back.style.display="none";
	front.style.display="block";

	if (window.widget)
		setTimeout ('widget.performTransition();', 0);
}


function showBack() {
	var front = document.getElementById("front");
	var back = document.getElementById("back");
        
	if (window.widget)
		widget.prepareForTransition("ToBack");
                
	front.style.display="none";
	back.style.display="block";
        
	if (window.widget)
		setTimeout ('widget.performTransition();', 0);  
}


function enterDone()
{
    document.images["doneButtonLeft"].src = "file:///System/Library/WidgetResources/button/glassbuttonleftclicked.png"
    document.images["doneButtonMiddle"].src = "file:///System/Library/WidgetResources/button/glassbuttonmiddleclicked.png"
    document.images["doneButtonRight"].src = "file:///System/Library/WidgetResources/button/glassbuttonrightclicked.png"
}
function exitDone()
{
    document.images["doneButtonLeft"].src = "file:///System/Library/WidgetResources/button/glassbuttonleft.png"
    document.images["doneButtonMiddle"].src = "file:///System/Library/WidgetResources/button/glassbuttonmiddle.png"
    document.images["doneButtonRight"].src = "file:///System/Library/WidgetResources/button/glassbuttonright.png"
}


function enterflip(event)
{
    document.getElementById('fliprollie').style.display = 'block';
    document.getElementById('flip').style.background = "url('file:///System/Library/WidgetResources/ibutton/white_i.png')";
}
function exitflip(event)
{
    document.getElementById('flip').style.background = "url('file:///System/Library/WidgetResources/ibutton/black_i.png')";
    document.getElementById('fliprollie').style.display = 'none';
}

var flipShown = false;
var animation = {duration:0, starttime:0, to:1.0, now:0.0, from:0.0, firstElement:null, timer:null};
function mousemove (event)
{
    if (!flipShown)
    {
        if (animation.timer != null)
        {
            clearInterval (animation.timer);
            animation.timer  = null;
        }
                
        var starttime = (new Date).getTime() - 13;
                
        animation.duration = 500;
        animation.starttime = starttime;
        animation.firstElement = document.getElementById ('flip');
        animation.timer = setInterval ("animate();", 13);
        animation.from = animation.now;
        animation.to = 1.0;
        animate();
        flipShown = true;
    }
}
function mouseexit (event)
{
    if (flipShown)
    {
        // fade in the info button
        if (animation.timer != null)
        {
            clearInterval (animation.timer);
            animation.timer  = null;
        }
                
        var starttime = (new Date).getTime() - 13;
                
        animation.duration = 500;
        animation.starttime = starttime;
        animation.firstElement = document.getElementById ('flip');
        animation.timer = setInterval ("animate();", 13);
        animation.from = animation.now;
        animation.to = 0.0;
        animate();
        flipShown = false;
    }
}
function animate()
{
    var T;
    var ease;
    var time = (new Date).getTime();
                
        
    T = limit_3(time-animation.starttime, 0, animation.duration);
        
    if (T >= animation.duration)
    {
        clearInterval (animation.timer);
        animation.timer = null;
        animation.now = animation.to;
    }
    else
    {
        ease = 0.5 - (0.5 * Math.cos(Math.PI * T / animation.duration));
        animation.now = computeNextFloat (animation.from, animation.to, ease);
    }
        
    animation.firstElement.style.opacity = animation.now;
}
function limit_3 (a, b, c)
{
    return a < b ? b : (a > c ? c : a);
}
function computeNextFloat (from, to, ease)
{
    return from + (to - from) * ease;
}
